@@ -43,8 +43,9 @@ import xml.dom.minidom
 _devnull = open(os.devnull, 'wb')
 
 # Cached information to speed up patch dependency checks
-latest_wine_commit  = None
-cached_patch_result = {}
+latest_wine_commit     = None
+latest_staging_version = None
+cached_patch_result    = {}
 
 class config(object):
     path_cache              = ".patchupdate.cache"
@@ -68,12 +69,14 @@ class PatchUpdaterError(RuntimeError):
 class PatchSet(object):
     def __init__(self, name, directory):
         self.name           = name
+        self.is_category    = False
         self.variable       = None
         self.directory      = directory
         self.fixes          = []
         self.changes        = []
         self.disabled       = False
         self.ifdefined      = None
+        self.categories     = []
 
         self.files          = []
         self.patches        = []
@@ -176,11 +179,13 @@ def _read_changelog():
             r = re.match("^([a-zA-Z0-9][^(]*)\((.*)\) ([^;]*)", line)
             if r: yield (r.group(1).strip(), r.group(2).strip(), r.group(3).strip())
 
-def _stable_compholio_version():
+def _latest_staging_version(only_stable=False):
     """Get version number of the latest stable release."""
     for package, version, distro in _read_changelog():
         if distro.lower() != "unreleased":
             return version
+        elif not only_stable:
+            return "%s (unreleased)" % version
 
 def _latest_wine_commit(commit=None):
     """Get latest wine commit."""
@@ -188,7 +193,7 @@ def _latest_wine_commit(commit=None):
         raise PatchUpdaterError("Please create a symlink to the wine repository in %s" % config.path_wine)
     if commit is None:
         commit = subprocess.check_output(["git", "rev-parse", "origin/master"], cwd=config.path_wine).strip()
-    assert len(commit) == 40
+    assert len(commit) == 40 and commit == commit.lower()
     return commit
 
 def enum_directories(revision, path):
@@ -230,6 +235,7 @@ def read_patchset(revision = None):
     all_patches = {}
     name_to_id  = {}
     all_bugids  = set()
+    categories  = {}
 
     # Read in sorted order (to ensure created Makefile doesn't change too much)
     for name, directory in sorted(enum_directories(revision, config.path_patches)):
@@ -294,6 +300,12 @@ def read_patchset(revision = None):
                     if i != j and any([fnmatch.fnmatch(f, val) for f in other_patch.modified_files]):
                         patch.auto_depends.add(j)
 
+            elif key == "category":
+                val = "category-%s" % val
+                if name_to_id.has_key(val):
+                    raise PatchUpdaterError("Category name in definition file %s collides with patchset %s" % (filename, val))
+                patch.categories.append(val)
+
             elif key == "fixes":
                 r = re.match("^[0-9]+$", val)
                 if r:
@@ -318,6 +330,24 @@ def read_patchset(revision = None):
             elif revision is None:
                 print "WARNING: Ignoring unknown command in definition file %s: %s" % (filename, line)
 
+        # If patch is not disabled then finally add it to the category
+        if not patch.disabled:
+            for category in patch.categories:
+                if not categories.has_key(category):
+                    categories[category] = set()
+                categories[category].add(i)
+
+
+    # Add virtual targets for all the categories
+    for category, indices in categories.iteritems():
+        patch = PatchSet(category, directory)
+        patch.is_category = True
+        patch.depends = indices
+
+        i = next(unique_id)
+        all_patches[i] = patch
+        name_to_id[name] = i
+
     # To simplify the task of keeping the bug list up-to-date, list all bugs
     # which might require attention.
     if revision is None:
@@ -543,38 +573,42 @@ def generate_script(all_patches):
             chunk_size = 20
             with progressbar.ProgressBar(desc=filename, total=2 ** len(indices) / chunk_size) as progress:
 
-                def test_apply(bitstring):
-                    set_apply = [(i, all_patches[i]) for u, i in zip(bitstring, indices) if u]
-                    set_skip  = [(i, all_patches[i]) for u, i in zip(bitstring, indices) if not u]
+                def test_apply(current):
+                    set_apply = [(i, all_patches[i]) for i in current]
+                    set_skip  = [(i, all_patches[i]) for i in indices if i not in current]
 
                     # Check if there is any patch2 which depends directly or indirectly on patch1.
                     # If this is the case we found an impossible situation, we can be skipped in this test.
                     for i, patch1 in set_apply:
                         for j, patch2 in set_skip:
                             if causal_time_smaller(patch2.verify_time, patch1.verify_time):
-                                return True # we can skip this test
+                                return None # we can skip this test
 
                     try:
                         original = original_content
                         for i, _ in set_apply:
                             original = patchutils.apply_patch(original, selected_patches[i][1], fuzz=0)
                     except patchutils.PatchApplyError:
-                        return False
+                        return current
 
-                    return True # everything is fine
+                    return None # everything is fine
 
-                def test_apply_seq(bitstrings):
-                    for bitstring in bitstrings:
-                        if not test_apply(bitstring):
-                            return False
-                    return True
+                def test_apply_seq(current_list):
+                    for current in current_list:
+                        failed = test_apply(current)
+                        if failed is not None:
+                            return failed
+                    return None
 
-                it = _split_seq(itertools.product([0,1], repeat=len(indices)), chunk_size)
-                for k, res in enumerate(pool.imap_unordered(test_apply_seq, it)):
-                    if not res:
+                iterables = []
+                for i in xrange(0, len(indices) + 1):
+                    iterables.append(itertools.combinations(indices, i))
+                it = _split_seq(itertools.chain(*iterables), chunk_size)
+                for k, failed in enumerate(pool.imap_unordered(test_apply_seq, it)):
+                    if failed is not None:
                         progress.finish("<failed to apply>")
                         raise PatchUpdaterError("Changes to file %s don't apply: %s" %
-                                                (filename, ", ".join([all_patches[i].name for i in indices])))
+                                                (filename, ", ".join([all_patches[i].name for i in failed])))
                     progress.update(k)
 
             # Update the dependency cache
@@ -590,11 +624,23 @@ def generate_script(all_patches):
     lines.append("patch_enable_all ()\n")
     lines.append("{\n")
     for i, patch in sorted([(i, all_patches[i]) for i in resolved], key=lambda x:x[1].name):
+        if patch.is_category: continue
         patch.variable = "enable_%s" % patch.name.replace("-","_").replace(".","_")
         lines.append("\t%s=\"$1\"\n" % patch.variable)
     lines.append("}\n")
     lines.append("\n")
-    lines.append("# Enable or disable a specific patchset\n")
+
+    lines.append("# Enable or disable all categories\n")
+    lines.append("category_enable_all ()\n")
+    lines.append("{\n")
+    for i, patch in sorted([(i, all_patches[i]) for i in resolved], key=lambda x:x[1].name):
+        if not patch.is_category: continue
+        patch.variable = "enable_%s" % patch.name.replace("-","_").replace(".","_")
+        lines.append("\t%s=\"$1\"\n" % patch.variable)
+    lines.append("}\n")
+    lines.append("\n")
+
+    lines.append("# Enable or disable a specific patchset/category\n")
     lines.append("patch_enable ()\n")
     lines.append("{\n")
     lines.append("\tcase \"$1\" in\n")
@@ -628,6 +674,11 @@ def generate_script(all_patches):
     # Generate code for applying all patchsets
     lines = []
     for i, patch in [(i, all_patches[i]) for i in resolved]:
+
+        # Categories do not have any files associated, so just skip over
+        if len(patch.files) == 0:
+            continue
+
         lines.append("# Patchset %s\n" % patch.name)
         lines.append("# |\n")
 
@@ -658,14 +709,16 @@ def generate_script(all_patches):
     with open(config.path_template_script) as template_fp:
         template = template_fp.read()
     with open(config.path_script, "w") as fp:
-        fp.write(template.format(patch_helpers="".join(lines_helpers).rstrip("\n"),
+        fp.write(template.format(latest_staging_version=_latest_staging_version(),
+                                 latest_wine_commit=latest_wine_commit,
+                                 patch_helpers="".join(lines_helpers).rstrip("\n"),
                                  patch_resolver="".join(lines_resolver).rstrip("\n"),
                                  patch_apply="".join(lines_apply).rstrip("\n")))
 
     # Add changes to git
     subprocess.call(["git", "add", config.path_script])
 
-def generate_markdown(all_patches, stable_patches, stable_compholio_version):
+def generate_markdown(all_patches, stable_patches):
     """Generate README.md including information about specific patches and bugfixes."""
 
     def _format_bug(mode, bugid, bugname):
@@ -685,11 +738,20 @@ def generate_markdown(all_patches, stable_patches, stable_compholio_version):
     # Compare with fixes for latest stable version
     for _, patch in stable_patches.iteritems():
         for bugid, bugname in patch.fixes:
-            key = bugid if bugid is not None else bugname
-            if all_fixes.has_key(key):
-                all_fixes[key][0] = 0
+            if bugid is not None and all_fixes.has_key(bugid):
+                all_fixes[bugid][0] = 0
+            elif all_fixes.has_key(bugname):
+                all_fixes[bugname][0] = 0
+            elif bugid is None:
+                for k, v in all_fixes.iteritems():
+                    if v[2] != bugname: continue
+                    if v[1] is None: continue
+                    all_fixes[v[1]][0] = 0
+                    break
+                else:
+                    all_fixes[bugname] = [-1, None, bugname]
             else:
-                all_fixes[key] = [-1, bugid, bugname]
+                all_fixes[bugid] = [-1, bugid, bugname]
 
     # Generate lists for all new and old fixes
     new_fixes = [(mode, bugid, bugname) for dummy, (mode, bugid, bugname) in
@@ -711,7 +773,7 @@ def generate_markdown(all_patches, stable_patches, stable_compholio_version):
             lines.append(_format_bug(mode, bugid, bugname))
         lines.append("")
         lines.append("")
-    lines.append("**Bug fixes and features in Wine Staging %s [%d]:**" % (stable_compholio_version, len(old_fixes)))
+    lines.append("**Bug fixes and features in Wine Staging %s [%d]:**" % (latest_staging_version, len(old_fixes)))
     lines.append("")
     lines.append("*Note: The following list only contains features and bug fixes which are not")
     lines.append("yet available in vanilla Wine. They are removed from the list as soon as they")
@@ -732,6 +794,25 @@ def generate_markdown(all_patches, stable_patches, stable_compholio_version):
     # Add changes to git
     subprocess.call(["git", "add", config.path_README_md])
 
+def wrap_changelog():
+
+    lines = []
+    with open(config.path_changelog) as fp:
+        for line in fp:
+            if line.startswith("  *") and len(line.rstrip("\n")) > 80:
+                wrapped = textwrap.wrap(line[3:].strip(), 80 - 4)
+                lines.append("  * %s\n" % "\n    ".join(wrapped))
+            else:
+                lines.append(line)
+
+    with open(config.path_changelog, "w") as fp:
+        for line in lines:
+            fp.write(line)
+
+    # Add changes to git
+    subprocess.call(["git", "add", config.path_changelog])
+
+
 if __name__ == "__main__":
 
     # Hack to avoid KeyboardInterrupts on different threads
@@ -740,19 +821,23 @@ if __name__ == "__main__":
         raise RuntimeError("CTRL+C pressed")
     signal.signal(signal.SIGINT, _sig_int)
 
+    tools_directory = os.path.dirname(os.path.realpath(__file__))
+    os.chdir(os.path.join(tools_directory, "./../.."))
+
     try:
 
-        # Get information about Wine and Compholio version
+        # Get information about Wine and Staging version
         latest_wine_commit       = _latest_wine_commit(sys.argv[1] if len(sys.argv) >= 2 else None)
-        stable_compholio_version = _stable_compholio_version()
+        latest_staging_version   = _latest_staging_version(only_stable=True)
 
         # Read current and stable patches
         all_patches    = read_patchset()
-        stable_patches = read_patchset(revision="v%s" % stable_compholio_version)
+        stable_patches = read_patchset(revision="v%s" % latest_staging_version)
 
         generate_ifdefined(all_patches)
         generate_script(all_patches)
-        generate_markdown(all_patches, stable_patches, stable_compholio_version)
+        generate_markdown(all_patches, stable_patches)
+        wrap_changelog()
 
     except PatchUpdaterError as e:
         print ""
